/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id: dns_query.cc 19925 Tue Feb 11 14:41:03 PST 2014 Zi Hu $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
   send DNS query over different protocols.
   by Zi Hu (zihu@usc.edu)
*/

#include <stdio.h>
#include <string.h>    //strlen
#include <stdlib.h>
#include <sys/socket.h>
#include <arpa/inet.h> //inet_addr , inet_ntoa , ntohs etc
#include <netinet/in.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <string.h>
#include <chrono>
#include <ctime>
#include <getopt.h>
#include "utility.h"
#include "types.h"
#include "mysocket.h"


#ifdef _GNUTLS_
#include <gnutls/gnutls.h>
#endif

#ifndef T_A
#define T_A 1
#endif

#include <sys/types.h>
#include <unistd.h>
#include "ssl_util.h"

#define TCP_IDLE_SOCKBUF_SIZE 100
#define DISABLE_NAGLE false
#define FEI_IP "128.9.160.190" // our servers
#define SUM_IP "128.9.160.224"
#define CAFILE "/etc/ssl/certs/ca-bundle.trust.crt"

using namespace std;
//for experiment purpose
std::chrono::time_point<std::chrono::system_clock> start_ts, end_ts, first_ts;
std::chrono::duration<double> elapsed_seconds;

//thread to receive response on TCP connection
pthread_t recv_thread;
//list<pthread_t> thread_pool;
//
#ifdef _GNUTLS_
//thread to receive response on TLS connection
pthread_t recv_thread_tls;
bool gnutls_init_done =false;
//list of tls connections
list<gnutls_session_t*> tls_conns;

#endif

//thread to receive response on TLS connection
pthread_t recv_thread_ssl;
//thread to receive responses on UDP socket
pthread_t recv_thread_udp;

//list of tcp connections
list<CSocket*> tcp_conns;
//list of udp sockets
list<CSocket*> udp_socks;

//list of ssl connections
list<SSL*> ssl_conns;

//message id, increase by 1 after each query.
unsigned short msg_id = rand()%65000;

//start time;
vector<chrono::time_point<chrono::system_clock>> st_ts_list;
//end time;
vector<chrono::time_point<chrono::system_clock>> ed_ts_list;

//record send/receive time. id->ts
map< unsigned short, chrono::time_point<chrono::system_clock> > st_ts_map;
map< unsigned short, chrono::time_point<chrono::system_clock> > ed_ts_map;

//list of DNS names for top websites from Alexa;
vector<string> dns_list;

//DNS resolver 
//resolver IP;
char dns_servers[MAX_BUF_SIZE];
//resolver hostname;
char resolver[MAX_BUF_SIZE];

//count the number of response 
int resp_count=0;
//disable nagle?
bool disable_nagle=false;
//the number of queries
int num_queries=0;

//delay between queries
int interval = 0;

//send queries over tls directly (without tls negotiation)
bool direct_tls = false;

bool SSL_init_done = false;

//enable TCP fastopen?
bool enable_tcpfast=false;
//validate server certificate?
bool validate_ca = false;

bool first_packet = true;

bool strict_tdns = false;

bool verbose_print = false;

//which protocol
string proto = "udp";

unsigned short dns_port = 53;

int parse_dns_response(unsigned char* buf, int buflen);
bool resolver_tls_check(CSocket &mysock);
bool create_tcp_connection(const char *host , unsigned int port, CSocket* sock);
//send DNS query over UDP
void resovle_dns_over_UDP (unsigned char* , int);
//send DNS query over TCP
void resolve_dns_over_TCP (unsigned char* , int);
//send DNS query over persistent TCP
void resolve_dns_over_persistent_TCP (unsigned char* , int);

void host_to_ip()
{
    struct hostent *hostent = gethostbyname(resolver);
    if (hostent == NULL) {
	    fprintf(stderr, "cannot resolve hostname: %s\n", resolver);
	    exit(1);
    };
    struct in_addr tmp;
    (void) memcpy((void*)&tmp.s_addr, (void*)(hostent->h_addr_list[0]), hostent->h_length);
    strcpy(dns_servers,inet_ntoa(tmp));
    //printf("host:%s, IP: %s, dns_servers:%s\n", resolver,  inet_ntoa(tmp), dns_servers);
}


#ifdef _GNUTLS_
static int _verify_certificate_callback(gnutls_session_t session)
{
    if(validate_ca)
    {
          unsigned int status;
          int ret;//, type;
          gnutls_certificate_type_t type;
          const char *hostname;
          gnutls_datum_t out;
          // read hostname 
          hostname = (char*) gnutls_session_get_ptr(session);
          printf("hostname:%s\n", hostname);
          // This verification function uses the trusted CAs in the credentials
          // structure. So you must have installed one or more CA certificates.
          ret = gnutls_certificate_verify_peers3(session, hostname, &status);
          if (ret < 0) {
              return GNUTLS_E_CERTIFICATE_ERROR;
          }
          type = gnutls_certificate_type_get(session);
          ret = gnutls_certificate_verification_status_print(status, type, &out, 0); 
          if (ret < 0) {
              return GNUTLS_E_CERTIFICATE_ERROR;
          }
          printf("err message: %s\n", out.data);
          gnutls_free(out.data);
          if (status != 0) 
            return GNUTLS_E_CERTIFICATE_ERROR;
          else
            return 0;
    }
    else
    {
        // notify gnutls to continue handshake normally
        return 0;
    }
}

//set up tls handshake
bool tls_handshake(gnutls_session_t* session, CSocket *mysock)
{
    const char* err_msg;
    //new tls session;
    gnutls_anon_client_credentials_t anoncred;
    gnutls_certificate_credentials_t xcred;
    gnutls_certificate_allocate_credentials(&xcred);
    gnutls_anon_allocate_client_credentials(&anoncred);

    //initialize x509 certificate 
    gnutls_certificate_set_x509_trust_file(xcred, CAFILE, GNUTLS_X509_FMT_PEM);
    gnutls_certificate_set_verify_function(xcred, _verify_certificate_callback);

    gnutls_init (session, GNUTLS_CLIENT);
    gnutls_session_set_ptr(*session, (void *) "sum.isi.edu");
    gnutls_server_name_set(*session, GNUTLS_NAME_DNS, "sum.isi.edu", strlen("sum.isi.edu"));
    int ret = gnutls_priority_set_direct(*session, "PERFORMANCE:NORMAL:EXPORT", &err_msg);
    if (ret < 0) 
    {
        if (ret == GNUTLS_E_INVALID_REQUEST) 
        {
            fprintf(stderr, "Syntax error at: %s\n", err_msg);
        }
        fprintf(stderr, "Syntax error at: %s\n", err_msg);
        exit(1);
    }
    gnutls_dh_set_prime_bits(*session, 512);
    gnutls_credentials_set(*session, GNUTLS_CRD_ANON, anoncred);
    gnutls_credentials_set(*session, GNUTLS_CRD_CERTIFICATE, xcred);
    gnutls_transport_set_ptr (*session, (gnutls_transport_ptr_t)(ssize_t)mysock->_sock);
    ret = gnutls_handshake(*session);
    if (ret != GNUTLS_E_SUCCESS) 
    {
        fprintf (stderr, "*** Handshake failed:%d, %s\n", ret, gnutls_strerror(ret));
        gnutls_perror(ret);
        return false;
    }
    return true;
}

//set up tls handshake without CA validation
bool tls_handshake_noca(gnutls_session_t* session, CSocket *mysock)
{
    //new tls session;
    gnutls_anon_client_credentials_t anoncred;
    //initialize TLS session
    gnutls_anon_allocate_client_credentials (&anoncred);
    gnutls_init (session, GNUTLS_CLIENT);
    gnutls_priority_set_direct(*session, "PERFORMANCE:+ANON-DH:!ARCFOUR-128", NULL);
    // put the anonymous credentials to the current session
    gnutls_credentials_set (*session, GNUTLS_CRD_ANON, anoncred);
    gnutls_transport_set_int (*session, mysock->_sock);
    gnutls_handshake_set_timeout (*session, GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);
    int ret = gnutls_handshake(*session);
    if (ret < 0)
    {
        fprintf (stderr, "*** Handshake failed\n");
        gnutls_perror(ret);
        return false;
    }
    else
    {
        return true;
    }
}

//thread to receive response on the tls connection
void* recv_response_tls_thread(void* args)
{
    gnutls_session_t *mysession = (gnutls_session_t *)args;
    char buf[2];
    memset(buf, 0, 2);
    char pkt_len_buf[2];
    memset(pkt_len_buf, 0, 2);
    char pkt_buf[MAX_BUF_SIZE];
    memset(pkt_buf, 0, MAX_BUF_SIZE);
    while(true)
    {
        unsigned short pkt_len=0;
        int toread = 0;
        int recv_bytes =0;
        //read first two bytes (the TCP DNS packet length);
        memset(pkt_len_buf, 0, 2);
        while(toread<2)
        {
            recv_bytes=gnutls_record_recv(*mysession, buf, 1);
            if(recv_bytes<=0)
            {
                return NULL;
            }
            if(recv_bytes==1)
            {
                pkt_len_buf[toread]=buf[0];
                toread++;
            }
        }

        pkt_len = ntohs(*((unsigned short*) pkt_len_buf));
        //read the rest packet;
        toread=0;
        memset(pkt_buf, 0, MAX_BUF_SIZE);
        while(toread<pkt_len)
        {
            recv_bytes=gnutls_record_recv(*mysession, buf, 1);
            if(recv_bytes<=0)
            {
                return NULL;
            }
            if(recv_bytes==1)
            {
                pkt_buf[toread]=buf[0];
                toread++;
            }
        }
        //parse the dns response
        parse_dns_response((unsigned char*)pkt_buf, pkt_len);
        if(resp_count==num_queries)
            break;
    }
}

//find one idle tls connection for existing ones. 
gnutls_session_t* get_idle_tlsconn(int &index)
{
    int i=0;
    if(tls_conns.size()>0)
    {
        index=i+1;
        return *(tls_conns.begin());
    }
    //no existing idle connection, need to create a new tls connection 
    CSocket *mysock = new CSocket;
    if(!create_tcp_connection(dns_servers, dns_port, mysock)) 
    {
        fprintf(stderr, "create tcp connection failed\n");
        exit(1);
    }
    //push it to the tcp connection list
    tcp_conns.push_back(mysock);
    if(!direct_tls)
    {
        //check if the resolver support TLS or not.
        if(!resolver_tls_check(*mysock))
        {
            fprintf(stderr, "resolver doesn't support tls\n");
            exit(1);
        }
    }
    //new tls session;
    gnutls_session_t* session = new gnutls_session_t;
    //tls handshake
    if(!tls_handshake(session, mysock))
    {
        fprintf(stderr, "tls handshake failed\n");
        exit(1);
    }
    //push it to the tcp connection list
    tls_conns.push_back(session);
    index=i+1;
    //creat a new thread to handle response from the socket;
    if(proto == "ptls")
    {
        int ret = pthread_create(&recv_thread_tls, NULL, recv_response_tls_thread, static_cast<void*>(session));
        if(ret!=0)
            fprintf(stderr, "create thread failed\n");
    }
    return session;
}

/*
 * do a "sticky" read: 
 * handle the fact that 
 */
int sticky_gnutls_read(gnutls_session_t session, unsigned char * buf, int buf_size)
{
	bool got_length = false;
	unsigned char *buf_cur = buf;
	int buf_size_left = buf_size;
	int recv_len;
	int dns_reply_len_needed;
	int dns_len;

	while (1) {
		if((recv_len = gnutls_record_recv(session, buf_cur, buf_size_left)) < 0) {
			cerr << "receive error: " << recv_len << endl;
			exit(1);
		};
		if (!got_length) {
			if (recv_len < 2) {
				cerr << "fragemented gnutls length not handled\n";
				exit(1);
			};
			dns_reply_len_needed = ntohs(*((unsigned short*) buf_cur)) + 2;
			got_length = true;
		};
		buf_cur += recv_len;
		buf_size_left -= recv_len;
		dns_reply_len_needed -= recv_len;
		if (dns_reply_len_needed <= 0)
			break;
	};
	return buf_cur - buf;  // reply length
}

// Perform a DNS query by sending a packet via TLS
void resolve_dns_over_TLS_w_sock(unsigned char *host , int query_type, CSocket &mysock)
{
    gnutls_session_t session;
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    int ret=0;
    int splen=0;
    int recv_len=0;

    if(!tls_handshake(&session, &mysock))
    {
        fprintf(stderr, "tls handshake failed\n");
        exit(1);
    }
    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);
    splen=0;
    //send the query 
    if((splen=gnutls_record_send(session, buf, plen))<0)
    {
        exit(1);
    }
    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
#if 0
    recv_len=0;
    if((recv_len=gnutls_record_recv(session, buf, MAX_BUF_SIZE))<0)
    {
        fprintf(stderr, "receive error:%d\n", recv_len);
        exit(1);
    }
#else
    recv_len = sticky_gnutls_read(session, buf, MAX_BUF_SIZE);
#endif /* 0 */
    parse_dns_response((unsigned char*)buf+2, recv_len-2);
    gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_CLOSE_NOTIFY);
    gnutls_bye(session, GNUTLS_SHUT_RDWR);
    gnutls_deinit(session);
    return;
}

// Perform a DNS query by sending a packet via TLS
void resolve_dns_over_TLS_EDNS(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);

    unsigned short qlen;
    unsigned short plen;
    int ret=0;
    int splen=0;
    int recv_len=0;
    //create a socket
    CSocket mysock;
    if(!mysock.create(SOCK_STREAM, IPPROTO_TCP))
    {
        exit(1);
    }
    if(!create_tcp_connection(dns_servers, dns_port, &mysock)) 
    {
        fprintf(stderr, "creat tcp connection failed\n");
        exit(1);
    }

    gnutls_global_init();
    if(!direct_tls)
    {
        //check if the resolver support TLS or not.
        if(!resolver_tls_check(mysock))
        {
            fprintf(stderr, "resolver doesn't support tls\n");
            exit(1);
        }
    }
    //start tls
    resolve_dns_over_TLS_w_sock(host, query_type, mysock);
    mysock.close();
    gnutls_global_deinit();
    first_packet=true;
    return;
}

// Perform a DNS query by sending a packet via persistent TLS 
void resolve_dns_over_pTLS_EDNS_nopipe(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    int splen=0;

    gnutls_session_t* session = NULL;
    int which_sock=0;

    //get idle connection;
    if(!gnutls_init_done)
    {
        gnutls_global_init();
        gnutls_init_done=true;
    }

    session = get_idle_tlsconn(which_sock);
    if(session == NULL)
    {
        fprintf(stderr, "failed to get idle tcp connection\n");
        exit(1);
    }

    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);


    splen=0;
    //send the query 
    if((splen=gnutls_record_send(*session, buf, plen))<0)
    {
        fprintf(stderr, "gnutls send:%d\n", splen);
        exit(1);
    }

    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    int recv_len=0;
    if((recv_len=gnutls_record_recv(*session, buf, MAX_BUF_SIZE))<0)
    {
        fprintf(stderr, "receive error:%d\n", recv_len);
        exit(1);
    }
    parse_dns_response((unsigned char*)buf+2, recv_len-2);

    return;
}


// Perform a DNS query by sending a packet via persistent TLS 
void resolve_dns_over_persistent_TLS(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    int splen=0;

    gnutls_session_t* session = NULL;
    int which_sock=0;

    //get idle connection;
    if(!gnutls_init_done)
    {
        gnutls_global_init();
        gnutls_init_done=true;
    }
    session = get_idle_tlsconn(which_sock);
    if(session == NULL)
    {
        fprintf(stderr, "failed to get idle tcp connection\n");
        exit(1);
    }

    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);


    splen=0;
    //send the query 
    if((splen=gnutls_record_send(*session, buf, plen))<0)
    {
        fprintf(stderr, "gnutls send:%d\n", splen);
        exit(1);
    }
    return;
}

void close_all_tls_conns()
{
    //close all tls connections;
    if(tls_conns.size()<=0)
        return;
    list<gnutls_session_t*>::iterator tit;
    for(tit=tls_conns.begin();tit!=tls_conns.end();tit++)
    {
        gnutls_session_t* session = *tit;
        gnutls_bye (*session, GNUTLS_SHUT_RDWR);
        gnutls_deinit (*session);
    }
    gnutls_global_deinit();
}
#endif

//parse response.
int parse_dns_response(unsigned char* buf, int buflen)
{
    unsigned short msgid = get_dns_msg_id(buf, buflen);
    resp_count++;
    end_ts = std::chrono::system_clock::now();
    ed_ts_map.insert(pair<unsigned short, chrono::time_point<chrono::system_clock> >(msgid, end_ts));
    //to print certain fields.
    if(verbose_print)
        print_dns_msg(buf, buflen);
    return 0;
}


//set up ssl handshake
SSL* ssl_handshake(CSocket *mysock)
{
    SSL_CTX* ctx = SSL_CTX_new (SSLv23_client_method ());
    if (!ctx) ERR_print_errors_fp (stderr);
    SSL *ssl = new SSL;
    ssl = (SSL*)outgoing_ssl_fd(ctx, mysock->_sock);
    if(!ssl) 
    {
        fprintf(stderr, "cannot create ssl\n");
        return NULL;
    }

    while(1) 
    {
        int r;
        ERR_clear_error();
        if( (r=SSL_do_handshake(ssl)) == 1)
            break;
        r = SSL_get_error(ssl, r); 
        if(r != SSL_ERROR_WANT_READ && r != SSL_ERROR_WANT_WRITE) 
        {
            fprintf(stderr, "could not ssl_handshake\n");
            return NULL;
        }
    }

    if(validate_ca) 
    {
        SSL_CTX_load_verify_locations(ctx, CAFILE, NULL);
        X509* x = SSL_get_peer_certificate(ssl);
        if(!x) 
        {
            fprintf(stderr, "SSL: no peer certificate\n");
        }
        else
        {
            //X509_print_fp(stdout, x); 
            long res=SSL_get_verify_result(ssl);
            if (res == X509_V_OK)
            {
                printf("validate success:%zu\n", res);
                X509_free(x);
            }
            else
            {
                printf("validate fail:%zu\n", res);
                X509_free(x);
                exit(1);
            }
        }
    }
    return ssl;
}


//receive response on TCP connection
int recv_tcp_response(CSocket *mysock, char* recv_buf, int buf_size)
{
    char buf[2];
    memset(buf, 0, 2);
    char pkt_len_buf[2];
    memset(pkt_len_buf, 0, 2);
    char pkt_buf[MAX_BUF_SIZE];
    memset(pkt_buf, 0, MAX_BUF_SIZE);
    unsigned short pkt_len=0;
    int toread = 0;
    int recv_bytes =0;
    //read first two bytes (the TCP DNS packet length);
    memset(pkt_len_buf, 0, 2);
    while(toread<2)
    {
        recv_bytes=mysock->recv_data_TCP((char*)buf, 1);
        if(recv_bytes<=0)
        {
            printf("receive bytes: %d\n", recv_bytes);
            return -1;
        }
        if(recv_bytes==1)
        {
            pkt_len_buf[toread]=buf[0];
            toread++;
        }
    }
    pkt_len = ntohs(*((unsigned short*) pkt_len_buf));
    if(pkt_len+2 > buf_size)
    {
        fprintf(stderr, "buf is not big enough to store the whole the tcp response");
        return -1;
    }

    memcpy(recv_buf, pkt_len_buf, 2);
    //read the rest packet;
    toread=0;
    memset(pkt_buf, 0, MAX_BUF_SIZE);
    while(toread<pkt_len)
    {
        recv_bytes=mysock->recv_data_TCP((char*)buf, 1);
        if(recv_bytes<=0)
        {
            //printf("receive bytes: %d\n", recv_bytes);
            return -1;
        }
        if(recv_bytes==1)
        {
            pkt_buf[toread]=buf[0];
            toread++;
        }
    }
    memcpy(recv_buf+2, pkt_buf, pkt_len);
    return pkt_len+2;
}

void fillin_dest_addr(struct sockaddr_in *dest)
{
    struct hostent *hostent = gethostbyname(dns_servers);
    if (hostent == NULL) {
	fprintf(stderr, "cannot resolve hostname: %s\n", dns_servers);
	exit(1);
    };

    dest->sin_family = hostent->h_addrtype;
    dest->sin_port = htons(dns_port);
    (void) memcpy((void*)&dest->sin_addr.s_addr, (void*)(hostent->h_addr_list[0]), hostent->h_length);
}


//starttls negotiation
bool resolver_tls_check(CSocket &mysock)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short plen;
    int splen=0;
    int recv_len=0;
    bool tls_bit_set = false;
    //construct a fake EDNS query packet; 
    //query name "STARTTLS", query class "CH", query type "TXT", with TLS bit set(the second bit in "Z").
    plen = format_edns_query(buf,(unsigned char*)"STARTTLS", ns_t_txt, ns_c_chaos, QT_TCP, 65530);
    //plen=construct_dns_query_EDNS(buf, MAX_BUF_SIZE, (const char*)"STARTTLS", ns_c_chaos, ns_t_txt, true);

    if(enable_tcpfast && first_packet)
    {
        struct sockaddr_in dest;
	fillin_dest_addr(&dest);
        //send the query with TCP fastopen. 
        if((splen=mysock.send_data_TCP_fo((char*)buf, plen, dest))<0)
        {
            exit(1);
        }
        first_packet=false;
    }
    else
    {
        //send the query via normal TCP.
        if((splen=mysock.send_data_TCP((char*)buf, plen))<0)
        {
            exit(1);
        }
    }

    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    if((recv_len = mysock.recv_data_TCP((char*)buf)) < 0) {
        fprintf(stderr, "receive error:%d\n", recv_len);
        exit(1);
    }

    //check if the tls bit is set in the response?
    tls_bit_set = is_tls_set(buf+2, recv_len-2);
    if (tls_bit_set) {
        return true;
    } else {
	if (strict_tdns) {
            printf("remote server doesn't support DNS-over-TLS\n");
            return false;
        } else {
            printf("remote server doesn't return TO bit, but non-strict\n");
            return true;
        };
    }
}


//thread to receive response on TCP connection
void* recv_response_thread(void* args)
{
    CSocket *mysock = (CSocket *)args;
    char pkt_buf[MAX_BUF_SIZE];
    while(mysock->is_alive())
    {
        memset(pkt_buf, 0, MAX_BUF_SIZE);
        int recv_len=0;
        if((recv_len=recv_tcp_response(mysock,(char*)pkt_buf, MAX_BUF_SIZE)) <0 )
        {
            fprintf(stderr, "receive error:%d\n", recv_len);
            exit(1);
        }
        //parse the dns response
        parse_dns_response((unsigned char*)pkt_buf+2, recv_len-2);
        if(resp_count==num_queries)
            break;
    }
    return NULL;
}


//thread to receive response on a UDP socket 
void* recv_response_thread_udp(void* args)
{
    struct sockaddr_in dest;
    fillin_dest_addr(&dest);
    CSocket *mysock = (CSocket *)args;
    //Receive the answer
    char buf[MAX_BUF_SIZE];
    while(mysock->is_alive())
    {
        memset(buf, 0, MAX_BUF_SIZE);
        int recv_len=0;
        if((recv_len=mysock->recv_data_UDP((char*)buf, dest))<0)
        {
            fprintf(stderr, "receive error:%d\n", recv_len);
            exit(1);
        }
        //parse the dns response
        parse_dns_response((unsigned char*)buf, recv_len);
        if(resp_count==num_queries)
            break;
    }
    return NULL;
}

//thread to receive response on the ssl connection
void* recv_response_ssl_thread(void* args)
{
    SSL *myssl = (SSL *)args;
    char buf[2];
    memset(buf, 0, 2);
    char pkt_len_buf[2];
    memset(pkt_len_buf, 0, 2);
    char pkt_buf[MAX_BUF_SIZE];
    memset(pkt_buf, 0, MAX_BUF_SIZE);
    while(true)
    {
        unsigned short pkt_len=0;
        memset(pkt_len_buf, 0, 2);
        if(SSL_read(myssl, (void*)&pkt_len, sizeof(pkt_len)) <= 0) 
        {
            fprintf(stderr, "ssl read error\n");
            return NULL;
        }
        pkt_len = ntohs(pkt_len);

        memset(pkt_buf, 0, MAX_BUF_SIZE);
        if(SSL_read(myssl, pkt_buf, pkt_len) <= 0)
        {
            fprintf(stderr, "ssl read error\n");
            return NULL;
        }

        //parse the dns response
        parse_dns_response((unsigned char*)pkt_buf, pkt_len);
        if(resp_count==num_queries)
            break;
    }
    return NULL;
}






// create a new tcp connection 
bool create_tcp_connection(const char *host , unsigned int port, CSocket* sock)
{
    if(!sock->create(SOCK_STREAM, IPPROTO_TCP))
    {
        return false;
    }
    if(disable_nagle)
    {
        sock->disable_nagle();
    }
    //enable fast open
    if(enable_tcpfast)
    {
        sock->enable_fastopen();
    }
    //normal handshake.
    else
    {
        if(!sock->connect(host, port))
        {
            sock->close();
            return false;
        }
    }
    return true;
}

//find one idle tcp connection from existing ones. 
CSocket* get_idle_tcpconn(int &index)
{
    int i=0;
    if(tcp_conns.size()>0)
    {
        index=i+1;
        return *(tcp_conns.begin());
    }
    //no existing idle connection, need to create a new tcp connection 
    CSocket *mysock = new CSocket;
    if(!create_tcp_connection(dns_servers, dns_port, mysock)) 
    {
        fprintf(stderr, "creat tcp connection failed\n");
        exit(1);
    }
    //push it to the tcp connection list
    tcp_conns.push_back(mysock);
    index=i+1;
    //creat a new thread to handle response from the socket;
    if(proto == "ptcp")
    {
        int ret = pthread_create(&recv_thread, NULL, recv_response_thread, static_cast<void*>(mysock));
        if(ret!=0)
            fprintf(stderr, "create thread failed\n");
    }
    return mysock;
}

//find one idle UDP socket from existing ones. 
CSocket* get_idle_udpsock(int &index)
{
    int i=0;
    if(udp_socks.size()>0)
    {
        index=i+1;
        return *(udp_socks.begin());
    }
    CSocket *mysock = new CSocket;
    if(!mysock->create(SOCK_DGRAM, IPPROTO_UDP))
    {
        fprintf(stderr, "creat UDP socket failed\n");
        exit(1);
    }
    //push it to the tcp connection list
    udp_socks.push_back(mysock);
    index=i+1;
    //creat a new thread to handle response from the socket;
    int ret = pthread_create(&recv_thread_udp, NULL, recv_response_thread_udp, static_cast<void*>(mysock));
    if(ret!=0)
        fprintf(stderr, "create thread failed\n");
    return mysock;
}


void resolve_dns_over_persistent_TCP(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);

    unsigned short plen;
    unsigned short qlen;
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;
    CSocket *mysock = NULL;
    int which_sock=0;
    //get idle connection;
    mysock = get_idle_tcpconn(which_sock);
    if(mysock == NULL)
    {
        fprintf(stderr, "failed to get idle tcp connection\n");
    }

    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);

    unsigned short splen=0;
    if(enable_tcpfast && first_packet)
    {
        struct sockaddr_in dest;
	fillin_dest_addr(&dest);
        //send the query with TCP fastopen. 
        if((splen=mysock->send_data_TCP_fo((char*)buf, plen, dest))<0)
        {
            exit(1);
        }
        first_packet=false;
    }
    else
    {
        //send the query via normal TCP.
        if((splen=mysock->send_data_TCP((char*)buf, plen))<0)
        {
            exit(1);
        }
    }
    return;
}

// Perform a DNS query by sending a packet via TCP 
void resolve_dns_over_pTCP_nopipe(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;
    CSocket *mysock = NULL;
    int which_sock=0;
    //get idle connection;
    mysock = get_idle_tcpconn(which_sock);
    if(mysock == NULL)
    {
        fprintf(stderr, "failed to get idle tcp connection\n");
    }

    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);

    unsigned short splen=0;
    if(enable_tcpfast && first_packet)
    {
        struct sockaddr_in dest;
	fillin_dest_addr(&dest);

        //send the query with TCP fastopen. 
        if((splen=mysock->send_data_TCP_fo((char*)buf, plen, dest))<0)
        {
            exit(1);
        }
        first_packet=false;
    }
    else
    {
        //send the query via normal TCP.
        if((splen=mysock->send_data_TCP((char*)buf, plen))<0)
        {
            exit(1);
        }
    }

    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    int recv_len=0;
    if((recv_len=recv_tcp_response(mysock,(char*)buf, MAX_BUF_SIZE)) <0 )
    {
        fprintf(stderr, "receive error:%d\n", recv_len);
        exit(1);
    }
    //parse the dns response;
    parse_dns_response(buf+2, recv_len-2 );

    return;
}


// Perform a DNS query by sending a packet via TCP with fastopen
void resolve_dns_over_TCP_fastopen(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);

    unsigned short qlen;
    unsigned short plen;
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;
    //create a socket
    CSocket mysock;
    if(!mysock.create(SOCK_STREAM, IPPROTO_TCP))
    {
        exit(1);
    }

    if(!create_tcp_connection(dns_servers, dns_port, &mysock)) 
    {
        fprintf(stderr, "creat tcp connection failed\n");
        exit(1);
    }
    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);

    struct sockaddr_in dest;
    fillin_dest_addr(&dest);

    int splen=0;
    //send the query 
    if((splen=mysock.send_data_TCP_fo((char*)buf, plen, dest))<0)
    {
        exit(1);
    }

    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    int recv_len=0;
    if((recv_len=recv_tcp_response(&mysock,(char*)buf, MAX_BUF_SIZE)) <0 )
    {
        fprintf(stderr, "receive error:%d\n", recv_len);
        exit(1);
    }
    //parse the dns response;
    parse_dns_response(buf+2, recv_len-2 );
    mysock.close();
    return;
}



// Perform a DNS query by sending a packet via TCP 
void resolve_dns_over_TCP(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);

    unsigned short qlen;
    unsigned short plen;

 
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;
    //create a socket
    CSocket mysock;
    if(!mysock.create(SOCK_STREAM, IPPROTO_TCP))
    {
        exit(1);
    }
    if(!create_tcp_connection(dns_servers, dns_port, &mysock)) 
    {
        fprintf(stderr, "creat tcp connection failed\n");
        exit(1);
    }
    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);

    int splen=0;
    if(enable_tcpfast)
    {
        struct sockaddr_in dest;
	fillin_dest_addr(&dest);
        //send the query with TCP fastopen. 
        if((splen=mysock.send_data_TCP_fo((char*)buf, plen, dest))<0)
        {
            exit(1);
        }
    }
    else
    {
        //send the query via normal TCP.
        if((splen=mysock.send_data_TCP((char*)buf, plen))<0)
        {
            exit(1);
        }
    }

    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    int recv_len=0;
    if((recv_len=recv_tcp_response(&mysock,(char*)buf, MAX_BUF_SIZE)) <0 )
    {
        fprintf(stderr, "receive error:%d\n", recv_len);
        exit(1);
    }
    //parse the dns response;
    parse_dns_response(buf+2, recv_len-2 );
    mysock.close();
    return;
}


//find one idle ssl connection for existing ones. 
SSL* get_idle_sslconn(int &index)
{
    int i=0;
    if(ssl_conns.size()>0)
    {
        index=i+1;
        return *(ssl_conns.begin());
    }
    //no existing idle connection, need to create a new ssl connection 
    CSocket *mysock = new CSocket;
    if(!create_tcp_connection(dns_servers, dns_port, mysock)) 
    {
        fprintf(stderr, "creat tcp connection failed\n");
        exit(1);
    }

    //push it to the tcp connection list
    tcp_conns.push_back(mysock);
    if(!direct_tls)
    {
        //check if the resolver support TLS or not.
        if(!resolver_tls_check(*mysock))
        {
          fprintf(stderr, "resolver doesn't support tls\n");
          exit(1);
        }
    }
    //new ssl session;
    SSL* ssl = NULL;
    if((ssl=ssl_handshake(mysock)) == NULL)
    {
        fprintf(stderr, "ssl handshake failed\n");
        exit(1);
    }
    //push it to the tcp connection list
    ssl_conns.push_back(ssl);
    index=i+1;
    //creat a new thread to handle response from the socket;
    if(proto == "pssl")
    {
        int ret = pthread_create(&recv_thread_ssl, NULL, recv_response_ssl_thread, static_cast<void*>(ssl));
        if(ret!=0)
            fprintf(stderr, "create thread failed\n");
    }
    return ssl;
}



// Perform a DNS query by sending a packet via SSL
void resolve_dns_over_SSL(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    int splen=0;

    //create a socket
    CSocket mysock;
    if(!mysock.create(SOCK_STREAM, IPPROTO_TCP))
    {
        exit(1);
    }
    if(!create_tcp_connection(dns_servers, dns_port, &mysock)) 
    {
        fprintf(stderr, "creat tcp connection failed\n");
        exit(1);
    }
    if(!direct_tls)
    {
        //check if the resolver support TLS or not.
        if(!resolver_tls_check(mysock))
        {
            fprintf(stderr, "resolver doesn't support tls\n");
            exit(1);
        }
    }
    //build SSL connection;
    SSL* myssl = NULL;
    if(!SSL_init_done)
    {
        ERR_load_SSL_strings();
        OpenSSL_add_all_algorithms();
        SSL_library_init();
        SSL_init_done = true;
    }

    if((myssl=ssl_handshake(&mysock)) == NULL)
    {
        fprintf(stderr, "ssl handshake failed\n");
        exit(1);
    }

    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);

    splen=0;
    //send the query 
    if((splen=SSL_write(myssl, buf, plen))<0)
    {
        fprintf(stderr, "SSL_write failed:%d\n", splen);
        exit(1);
    }
    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short pkt_len=0;
    if(SSL_read(myssl, (void*)&pkt_len, sizeof(pkt_len)) <= 0) 
    {
        fprintf(stderr, "ssl read error\n");
        exit(1);
    }
    pkt_len = ntohs(pkt_len);
    memset(buf, 0, MAX_BUF_SIZE);
    if(SSL_read(myssl, buf, pkt_len) <= 0)
    {
        fprintf(stderr, "ssl read error\n");
        exit(1);
    }
    parse_dns_response((unsigned char*)buf, pkt_len);
    SSL_shutdown(myssl);
    SSL_shutdown(myssl);
    SSL_free(myssl);
    mysock.close();
    first_packet=true;
    return;
}


// Perform a DNS query by sending a packet via persistent SSL
void resolve_dns_over_pSSL_nopipe(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    int splen=0;

    SSL* myssl = NULL;
    int which_sock=0;

    //get idle connection;
    if(!SSL_init_done)
    {
        ERR_load_SSL_strings();
        OpenSSL_add_all_algorithms();
        SSL_library_init();
        SSL_init_done = true;
    }

    myssl = get_idle_sslconn(which_sock);
    if(myssl == NULL)
    {
        fprintf(stderr, "failed to get idle ssl connection\n");
        exit(1);
    }

    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);


    splen=0;
    //send the query 
    if((splen=SSL_write(myssl, buf, plen))<0)
    {
        fprintf(stderr, "SSL_write failed:%d\n", splen);
        exit(1);
    }

    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short pkt_len=0;
    if(SSL_read(myssl, (void*)&pkt_len, sizeof(pkt_len)) <= 0) 
    {
        fprintf(stderr, "ssl read error\n");
        exit(1);
    }
    pkt_len = ntohs(pkt_len);
    memset(buf, 0, MAX_BUF_SIZE);
    if(SSL_read(myssl, buf, pkt_len) <= 0)
    {
        fprintf(stderr, "ssl read error\n");
        exit(1);
    }
    parse_dns_response((unsigned char*)buf, pkt_len);
    return;
}


// Perform a DNS query by sending a packet via persistent SSL
void resolve_dns_over_persistent_SSL(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    int splen=0;

    SSL* myssl = NULL;
    int which_sock=0;

    //get idle connection;
    if(!SSL_init_done)
    {
        ERR_load_SSL_strings();
        OpenSSL_add_all_algorithms();
        SSL_library_init();
        SSL_init_done=true;
    }

    myssl = get_idle_sslconn(which_sock);
    if(myssl == NULL)
    {
        fprintf(stderr, "failed to get idle ssl connection\n");
        exit(1);
    }

    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_TCP, msg_id++);

    splen=0;
    //send the query 
    if((splen=SSL_write(myssl, buf, plen))<0)
    {
        fprintf(stderr, "SSL_write failed:%d\n", splen);
        exit(1);
    }
    return;
}



// Perform a DNS query by sending a packet via UDP
void resolve_dns_over_UDP(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;
    //create a socket
    CSocket mysock;
    if(!mysock.create(SOCK_DGRAM, IPPROTO_UDP))
    {
        exit(1);
    }
    struct sockaddr_in dest;
    fillin_dest_addr(&dest);
    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_UDP, msg_id++);


    int splen=0;
    //send the query 
    if((splen=mysock.send_data_UDP((char*)buf, plen, dest))<0)
    {
        exit(1);
    }

    //Receive the answer
    memset(buf, 0, MAX_BUF_SIZE);
    int recv_len=0;
    if((recv_len=mysock.recv_data_UDP((char*)buf, dest))<0)
    {
        fprintf(stderr, "receive error:%d\n", recv_len);
        exit(1);
    }
    //parse the dns response;
    parse_dns_response(buf,recv_len);
    mysock.close();
    return;
}

// Perform a DNS query by sending packets via UDP in parallel
void resolve_dns_over_concurrent_UDP(unsigned char *host , int query_type)
{
    unsigned char buf[MAX_BUF_SIZE];
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned short qlen;
    unsigned short plen;
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;
    CSocket *mysock = NULL;
    int which_sock=0;
    //get idle connection;
    mysock = get_idle_udpsock(which_sock);
    if(mysock == NULL) {
        fprintf(stderr, "failed to get idle tcp connection\n");
	exit(1);
    }

    struct sockaddr_in dest;
    fillin_dest_addr(&dest);
    //construct a DNS query packet;
    plen=format_dns_query(buf, qlen, host, query_type, ns_c_in, QT_UDP, msg_id++);
    int splen=0;
    //send the query 
    if((splen=mysock->send_data_UDP((char*)buf, plen, dest))<0)
    {
        exit(1);
    }
    return;
}




//usage information
void usage(char* prog)
{
    cout << "usage: " << prog  << "\n"
    << "-f/--dnslist-file = file, file contains a list of DNS names \n"
    << "-r/--dns-resolver = resolver ip, dns resolver ip \n"
    << "-d/--disable-nagle, disable nagle \n"
    << "-o/--enable-tcpfast, enable TCP fastopen \n"
    << "-n/--num-queries = number, total number of queries to send \n"
    << "-p/--port-number = port, dns server port number \n" 
    << "-V/--verbose, print details of responses (if not specified, we only print latency stats) \n"
#ifdef _GNUTLS_
    << "-t/--protocol = protocol, could be tcp/udp/tls/ssl/ptcp/cudp/ptls/pssl/ptcpnp/ptlsnp/psslnp \n"
    << "-s/--strict: strict TLS mode (requires TO bit returned); default is relaxed\n"
    << "-e/--direct-tls, send query over tls directly, without negotiation. by default\n"
    << "-v/--validate-ca, validate Certificate \n"
#else
    << "-t/--protocol = protocol, could be tcp/udp/ssl/ptcp/cudp/pssl/ptcpnp/psslnp\n"
    << "-e/--direct-tls, send query over ssl directly, without negotiation. by default\n"
#endif
    << "-i/--interval = ms, time interval between queries\n"
    << "-h/--help, show this message\n"
    << endl
    << "So an example: ./digit -t pssl -f /tmp/names -r 128.9.160.190 -p 10053\n\n";
}

// parse arguments
int
options(int argc, char *argv[])
{
    static struct option longopts[]={   
          {"dnslist-file", required_argument, NULL, 'f'}, 
          {"disable-nagle", no_argument, NULL, 'd'}, 
          {"enable-tcpfast", no_argument, NULL, 'o'}, 
          {"validate-ca", no_argument, NULL, 'v'}, 
          {"direct-tls", no_argument, NULL, 'e'}, 
          {"dns-resolver", required_argument, NULL, 'r'}, 
          {"strict", no_argument, NULL, 's'}, 
          {"protocol", required_argument, NULL, 't'}, 
          {"num-queries", required_argument, NULL, 'n'}, 
          {"port-number", required_argument, NULL, 'p'}, 
          {"verbose", no_argument, NULL, 'V'}, 
          {"interval", required_argument, NULL, 'i'}, 
          {"help", no_argument, NULL, 'h'}, 
          {NULL, 0, NULL, 0}
    };  

    int ch; 
    while ((ch = getopt_long(argc, argv, "f:r:n:p:i:t:dsceovVh", longopts, NULL)) != -1) {
        switch (ch) {
        case 'f':
            read_dnslist(optarg, dns_list);
            break;
        case 't':
            proto=optarg;
            break;
        case 'd':
            disable_nagle=true;
            break;
        case 'o':
            enable_tcpfast=true;
            break;
        case 'v':
            validate_ca=true;
            break;
        case 'V':
            verbose_print=true;
            break;
        case 'e':
            direct_tls=true;
            break;
        case 'n':
            num_queries=atoi(optarg);
            break;
        case 'i':
            interval=atoi(optarg);
            break;
        case 'p':
            dns_port=atoi(optarg);
            break;
        case 's':
	        strict_tdns = true;
            break;
        case 'r':
	        strncpy(resolver, optarg, MAX_BUF_SIZE);
	        resolver[MAX_BUF_SIZE-1] = 0; // null terminate
            host_to_ip();
            break;
        case 'h':
            usage(argv[0]);
            exit(1);
        default:
            usage(argv[0]);
            exit(1);
        }
    }
    return optind;
}

void close_all_socks()
{
    list<CSocket*>::iterator tit;
    //close all the tcp connections;
    if(tcp_conns.size()>0 )
    {
        for(tit=tcp_conns.begin();tit!=tcp_conns.end();tit++)
        {
            (*tit)->close();
        }
    }
    if(udp_socks.size()>0)
    {
        for(tit=udp_socks.begin();tit!=udp_socks.end();tit++)
        {
            (*tit)->close();
        }
    }
}


void close_all_ssl_conns()
{
    //close all tls connections;
    if(ssl_conns.size()<=0)
        return;
    list<SSL*>::iterator tit;
    for(tit=ssl_conns.begin();tit!=ssl_conns.end();tit++)
    {
        SSL* myssl = *tit;
        SSL_shutdown(myssl);
        SSL_free(myssl);
    }
}


void compute_consumed_time()
{
    chrono::time_point<chrono::system_clock> last_ts = first_ts;
    double completion_t=0.0;
    double sum_t=0.0;
    int i=0;
    map< unsigned short, chrono::time_point<chrono::system_clock> >::iterator sit;
    map< unsigned short, chrono::time_point<chrono::system_clock> >::iterator eit;
    printf("#fsdb\tindex\tt_complete\tt_avg\tt_individual\tt_sum\tt_mean\tid\tquery_send_ts\tresponse_receive_ts\tprogram_start_ts\n");
    for(sit=st_ts_map.begin(); sit!=st_ts_map.end(); sit++, i++)
    {
        eit=ed_ts_map.find(sit->first);
        if(eit==ed_ts_map.end())
        {
            fprintf(stderr, "no match for query for: %d\n", i+1);
            continue;
        }
        else
        {
            if(eit->second <= last_ts)
            {
                elapsed_seconds = last_ts - first_ts;
            }
            else
            {
                elapsed_seconds = eit->second - first_ts;
                last_ts = eit->second;
            }

            completion_t = elapsed_seconds.count();
            std::chrono::duration<double> st_dur, rt_dur, ft_dur;
            st_dur = (sit->second).time_since_epoch();
            rt_dur = (eit->second).time_since_epoch();
            ft_dur = (first_ts).time_since_epoch();
            elapsed_seconds = eit->second - sit->second;
            sum_t += elapsed_seconds.count();
            printf("%d\t%f\t%f\t%f\t%f\t%f\t%d\t%f\t%f\t%f\n", \
                i+1, completion_t, completion_t/(i+1), elapsed_seconds.count(), sum_t, sum_t/(i+1), sit->first, st_dur.count(), rt_dur.count(), ft_dur.count());
        }
    }
}

//handle ctrl+c
void sig_handler (int sig)
{
    compute_consumed_time();
    fflush(stdout);
    close_all_socks();
#ifdef _GNUTLS_
    close_all_tls_conns();
#endif
    close_all_ssl_conns();
    exit(1);
}

void send_query (int index)
{
    int j=0;
    if(index>=num_queries)
    {
        if(proto=="ptcp")
            pthread_join(recv_thread, NULL);
        else if(proto=="cudp")
            pthread_join(recv_thread_udp, NULL);
#ifdef _GNUTLS_
        else if(proto=="ptls")
            pthread_join(recv_thread_tls, NULL);
#endif
        else if(proto=="pssl")
            pthread_join(recv_thread_ssl, NULL);
        compute_consumed_time();
        fflush(stdout);
        close_all_socks();
#ifdef _GNUTLS_ 
        close_all_tls_conns();
#endif
        close_all_ssl_conns();
        exit(1);
    }
    else
    {
        if(index>=dns_list.size())
        {
          //printf("exceed max: %d\t%lu\n", i, dns_list.size());
          j=index%dns_list.size();
        }
        else
        {
          j=index;
        }

        start_ts = std::chrono::system_clock::now();
        st_ts_map.insert(pair<unsigned short, chrono::time_point<chrono::system_clock> >(msg_id, start_ts));
        if(proto=="ptcp")
            resolve_dns_over_persistent_TCP((unsigned char*)dns_list[j].c_str(), T_A);
        else if(proto=="cudp")
            resolve_dns_over_concurrent_UDP((unsigned char*)dns_list[j].c_str(), T_A);
        else if(proto=="pssl")
            resolve_dns_over_persistent_SSL((unsigned char*)dns_list[j].c_str(), T_A);
        else if(proto=="tcp")
            //resolve_dns_over_TCP_fastopen((unsigned char*)dns_list[j].c_str() , T_A);
            resolve_dns_over_TCP((unsigned char*)dns_list[j].c_str() , T_A);
        else if(proto=="udp")
            resolve_dns_over_UDP((unsigned char*)dns_list[j].c_str() , T_A);
#ifdef _GNUTLS_
        else if(proto=="tls")
            resolve_dns_over_TLS_EDNS((unsigned char*)dns_list[j].c_str() , T_A);
        else if(proto=="ptls")
            resolve_dns_over_persistent_TLS((unsigned char*)dns_list[j].c_str(), T_A);
        else if(proto=="ptlsnp")
            resolve_dns_over_pTLS_EDNS_nopipe((unsigned char*)dns_list[j].c_str() , T_A);
#endif
        else if(proto=="ssl")
            resolve_dns_over_SSL((unsigned char*)dns_list[j].c_str() , T_A);
        else if(proto=="ptcpnp")
            resolve_dns_over_pTCP_nopipe((unsigned char*)dns_list[j].c_str() , T_A);
        else if(proto=="psslnp")
            resolve_dns_over_pSSL_nopipe((unsigned char*)dns_list[j].c_str() , T_A);
        else
        {
            fprintf(stderr, "unknown protocol\n");
            exit(1);
        }
    }
}

//test latency of persistent TCP by sending many queries;
void latency_evaluation(int num_queries)
{
    for(int i=0; i<=num_queries; i++)
    {
        send_query(i);
    }
}



int qindex = 0;
void timer_handler (int sig)
{
    send_query(qindex);
    qindex++;
}

//main function
int main( int argc , char *argv[])
{
    signal(SIGQUIT, sig_handler);
    signal(SIGINT, sig_handler);

    if(argc<2)
    {
        usage(argv[0]);
        exit(1);
    }

    int i=0;
    //parse parameters;
    i = options(argc, argv);
    argc -= i;
    argv += i;
    if(dns_list.size()==0)
    {
        printf("no domain names in the file\n");
        return 1;
    }

    if(num_queries == 0)
        num_queries = dns_list.size();
    struct sigaction sa; 
    struct itimerval timer ;
    memset ( &sa, 0, sizeof ( sa ) ) ; 
    sa.sa_handler = &timer_handler ;
    sigaction ( SIGALRM, &sa, NULL );
    int ts_sec = interval/1000;
    int ts_usec = (interval%1000)*1000;
    timer.it_value.tv_sec = ts_sec ; 
    timer.it_value.tv_usec = ts_usec;
    timer.it_interval.tv_sec = ts_sec;
    timer.it_interval.tv_usec = ts_usec;

    //do DNS query;
    first_ts = std::chrono::system_clock::now();
    if(interval==0)
    {
        latency_evaluation(num_queries);
        compute_consumed_time();
        fflush(stdout);
        close_all_socks();
#ifdef _GNUTLS_ 
        close_all_tls_conns();
#endif
        close_all_ssl_conns();
    }
    else
    {
        setitimer ( ITIMER_REAL, &timer, NULL ) ; 
        while(true);
    }
    return 0;
}
